import React, { useEffect, useRef, useState } from 'react';
import '@mescius/wijmo.styles/wijmo.css';
import { MultiRow, MultiRowCellGroup, MultiRowCell, MultiRowCellTemplate } from '@mescius/wijmo.react.grid.multirow';
import { InputDate } from '@mescius/wijmo.react.input';
import * as wjMultirow from '@mescius/wijmo.grid.multirow';
import * as wjCore from '@mescius/wijmo';
import { getData, CarData, MessageSeverity } from './data';
import { io } from "socket.io-client";

const socket = io("http://localhost:9001", { autoConnect: false });

function App() {
    const [data, setData] = useState<CarData[]>([]);
    const [originalData, setOriginalData] = useState<CarData[]>([]);
    const [validationMessageShowing, setValidationMessageShowing] = useState(false);
    const gridRef = useRef<MultiRow>(null);
    // https://medium.com/geographit/accessing-react-state-in-event-listeners-with-usestate-and-useref-hooks-8cceee73c559
    const originalDataRef = useRef<CarData[]>();

    useEffect(() => {
        const theData = getData();
        // Because data in the array contains objects, we need to create a new array with new objects
        setData(theData.map((item: any) => ({ ...item })));
        // Stores the state of the data when it was first loaded
        setOriginalData(theData.map((item: any) => ({ ...item })));
        originalDataRef.current = theData.map((item: any) => ({ ...item }));
    }, []);

    useEffect(() => {
        // no-op if the socket is already connected
        socket.connect();

        return () => {
            socket.disconnect();
        };
    }, []);

    useEffect(() => {
        // Listen for changed data from server
        socket.on("received_changed_data", (receivedData: any) => {
            console.log('Received Changed Data: ', receivedData);

            if (receivedData.changedRows && receivedData.changedRows.length > 0) {

                // Loop through changedRows
                for (let changedRow of receivedData.changedRows) {
                    // Find the index of  the corresponding record in originalDataRef
                    let index = originalDataRef.current?.findIndex((row: any) => row.carNumber === changedRow.carNumber);

                    // If the record is found, replace it
                    if (originalDataRef.current && index && index !== -1) {
                        // Convert date strings to Date objects
                        changedRow.inboundDate = new Date(changedRow.inboundDate);
                        originalDataRef.current![index] = { ...changedRow };
                    }
                }

                // Set Data equal to newData
                console.log('SET DATA - START!', originalDataRef.current);
                // Setting the state variable "data" will cause the grid to re-render
                setData(originalDataRef.current!.map((item: any) => ({ ...item })));
                // Replace the originalData with the new data
                setOriginalData(originalDataRef.current!.map((item: any) => ({ ...item })));
                console.log('SET DATA - COMPLETE!');
            }
        });

        return () => {
            socket.off("received_changed_data");
        }
    }, []);

    // Gather rows that have been changed
    const saveChanges = () => {
        const changedRows = [];

        for (let i = 0; i < gridRef.current?.control.itemsSource.length; i++) {
            const gridItem = gridRef.current?.control.itemsSource[i];
            // Is there a better way to determine if data in a row has been edited?
            if (originalData && JSON.stringify(gridItem) !== JSON.stringify(originalData[i])) {
                changedRows.push(gridItem);
            }
        }

        console.log('Changed Rows: ', changedRows);
        // Send changes to the server in open socket
        socket.emit("send_changed_data", { changedRows });
    };

    const changeValidationMessageShowing = () => {
        let newData = data.map((item: any) => ({ ...item }));
        let car6 = newData.find((item: any) => item.id === 6);
        if (car6) {
            if (!validationMessageShowing) {
                setValidationMessageShowing(true);
                car6.message = "This is a validation error message";
                car6.severity = MessageSeverity.Error
            } else {
                setValidationMessageShowing(false);
                car6.message = "";
            }
            setData((data: any) => {
                newData.forEach((newItem) => {
                    let item = data.find((t: any) => t.id == newItem.id);
                    if (item) {
                        // remove any extra properties
                        for (let prop in item) {
                            if (!newItem[prop]) {
                                delete item[prop];
                            }
                        }
                        // copy data to current items
                        for (let prop in newItem) {
                            item[prop] = newItem[prop];
                        }
                    }
                });
                gridRef.current?.control.collectionView.refresh();
                return data;
            });
        }
    }

    const MessageCellTemplate: React.FC<{ context: any, gridRef: any }> = ({ context, gridRef }) => {

        const getColor = (severity?: MessageSeverity) => {
            switch (severity) {
                case MessageSeverity.Error:
                    return 'red';
                case MessageSeverity.Warning:
                    return 'orange';
                case MessageSeverity.Success:
                    return 'green';
                case MessageSeverity.Info:
                    return 'blue';
                default:
                    return 'black';
            }
        };

        // useEffect does not fix the flushSync error
        // useEffect(() => {
        //     if (context.item.message) {
        //         context.row.height = null;
        //     } else {
        //         context.row.height = 0
        //     }
        // }, [context.item.message]);

        // console.log('Rendering Message for Car Number: ', context.item.carNumber, ' Index: ', context.row.index, ' Message: ', context.item.message);

        // This code works on first render and when showing/hiding the validation message.
        // However, after receiving new data in the web socket and setting the "data" state variable
        // we get the following error in the console:
        //      flushSync was called from inside a lifecycle method
        // The errer is not called bye this code. The stack trace says it comes from the MultiRow control.
        // If the line "context.row.height = 0" is commented out, the error does not occur.
        if (context.item.message) {
            // We want to show the "Message" line if a message exists
            context.row.height = null;
        } else {
            // Hide the "Message" line if no message exists
            // Commenting out this line will cause the error to not occur
            // context.row.height = 0
        }
        return context.item.message
            ? <span style={{ color: getColor(context.item.severity) }}>
                {context.item.message}
            </span>
            : null;
    }

    function loadedRows(s: wjMultirow.MultiRow, e: wjCore.EventArgs) {
        // hides the Feedback rows that doesn't have a message to display
        s.rows.forEach((row: any) => {
            let item = row.dataItem;
            if (row.recordIndex === 1 && !item.message) {
                row.visible = false;
            } else {
                row.visible = true;
            }
        })
    }

    return (
        <div className="container-fluid">
            <h1>Valid Cars - Multi Row</h1>
            <button onClick={() => saveChanges()}>Save</button>
            <button onClick={() => changeValidationMessageShowing()}>Change Validation Message Showing</button>

            <MultiRow ref={gridRef} itemsSource={data} loadedRows={loadedRows}>
                <MultiRowCellGroup>
                    <MultiRowCell binding="carNumber" header="Car Number" width={150} isReadOnly={true} />
                    <MultiRowCell binding="inboundDate" header="Inbound Date" width={150} cellType="Cell">
                        <MultiRowCellTemplate cellType="CellEdit" template=
                            {(context: any) => (
                                <InputDate value={context.value}
                                    valueChanged={(inputDate: any) => context.value = inputDate.value} />
                            )} />
                    </MultiRowCell>
                    <MultiRowCell binding="shopProcessing" header="Shop Processing" width={150} />
                    <MultiRowCell binding="message" header="Feedback Message" colspan={3} isReadOnly={true} allowSorting={false}>
                        <MultiRowCellTemplate cellType="Cell" template=
                            {(context: any) => <MessageCellTemplate context={context} gridRef={gridRef} />} />
                    </MultiRowCell>
                </MultiRowCellGroup>
            </MultiRow>
        </div>
    );
}

export default App;
